﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using HeuristicLab.Common;
using HeuristicLab.Core;

namespace GradientBoostedTrees {
  public class LogisticRegressionLoss : ILossFunction {
    public double GetLoss(IEnumerable<double> target, IEnumerable<double> pred, IEnumerable<double> weight) {
      var targetEnum = target.GetEnumerator();
      var predEnum = pred.GetEnumerator();
      var weightEnum = weight.GetEnumerator();

      double s = 0;
      while (targetEnum.MoveNext() & predEnum.MoveNext() & weightEnum.MoveNext()) {
        // assert target == 0 or target == 1
        if (!targetEnum.Current.IsAlmost(0.0) && !targetEnum.Current.IsAlmost(1.0))
          throw new NotSupportedException("labels must be 0 or 1 for logistic regression loss");
        double f = Math.Max(-7, Math.Min(7, predEnum.Current)); // threshold for exponent
        var probPos = Math.Exp(2 * f) / (1 + Math.Exp(2 * f));
        s += weightEnum.Current * (-targetEnum.Current * Math.Log(probPos) - (1 - targetEnum.Current) * Math.Log(1 - probPos));
      }
      if (targetEnum.MoveNext() | predEnum.MoveNext() | weightEnum.MoveNext())
        throw new ArgumentException("target, pred and weight have differing lengths");

      return s;
    }

    public IEnumerable<double> GetLossGradient(IEnumerable<double> target, IEnumerable<double> pred, IEnumerable<double> weight) {
      var targetEnum = target.GetEnumerator();
      var predEnum = pred.GetEnumerator();
      var weightEnum = weight.GetEnumerator();

      while (targetEnum.MoveNext() & predEnum.MoveNext() & weightEnum.MoveNext()) {
        // assert target == 0 or target == 1
        if (!targetEnum.Current.IsAlmost(0.0) && !targetEnum.Current.IsAlmost(1.0))
          throw new NotSupportedException("labels must be 0 or 1 for logistic regression loss");
        double f = Math.Max(-7, Math.Min(7, predEnum.Current)); // threshold for exponent
        var probPos = Math.Exp(2 * f) / (1 + Math.Exp(2 * f));
        yield return weightEnum.Current * (targetEnum.Current - probPos) / (probPos * probPos - probPos);
      }
      if (targetEnum.MoveNext() | predEnum.MoveNext() | weightEnum.MoveNext())
        throw new ArgumentException("target, pred and weight have differing lengths");
    }

    public LineSearchFunc GetLineSearchFunc(IEnumerable<double> target, IEnumerable<double> pred, IEnumerable<double> weight) {
      var targetArr = target.ToArray();
      var predArr = pred.ToArray();
      var weightArr = weight.ToArray();
      // weights are not supported yet
      // when weights are supported we need to calculate a weighted median
      Debug.Assert(weightArr.All(w => w.IsAlmost(1.0)));

      if (targetArr.Length != predArr.Length || predArr.Length != weightArr.Length)
        throw new ArgumentException("target, pred and weight have differing lengths");

      // line search for abs error
      LineSearchFunc lineSearch = (idx, startIdx, endIdx) => {
        double sumY = 0.0;
        double sumDiff = 0.0;
        for (int i = startIdx; i <= endIdx; i++) {
          var yi = (targetArr[idx[i]] - predArr[idx[i]]);
          var wi = weightArr[idx[i]];

          sumY += wi * yi;
          sumDiff += wi * Math.Abs(yi) * (1 - Math.Abs(yi));

        }
        // prevent divByZero
        sumDiff = Math.Max(1E-12, sumDiff);
        return 0.5 * sumY / sumDiff;
      };
      return lineSearch;

    }

    public override string ToString() {
      return "Logistic regression loss";
    }
  }
}
